"""
Create Entity order:
    1. Create Entity
    2. Create Attribute, Group
    3. Recreate Entity's view

Update Entity order:
    1. Remove Entity's view
    2. Create Entity
    3. Create Attribute, Group
    4. Recreate Entity's view

Remove Entity order:
    1. Remove Entity's view
    2. Remove Attribute, Group
    3. Remove Entity
"""

import pydantic

from .link import Link
from .group import Group
from .entity import Entity
from .priority import Priority
from .attribute import Attribute


class Migration(pydantic.BaseModel):
    priority: Priority


class CreateEntity(Migration):
    priority: Priority = Priority.FIRST
    entity: Entity

    def __str__(self):
        return f"<CreateEntity '{self.entity.name}'>"


class RemoveEntity(Migration):
    priority: Priority = Priority.THIRD
    entity: Entity

    def __str__(self):
        return f"<RemoveEntity '{self.entity.name}'>"


class RecreateEntityView(Migration):
    priority: Priority = Priority.THIRD
    entity: Entity

    def __str__(self):
        return f"<RecreateEntityView '{self.entity.name}'>"


class RemoveEntityView(Migration):
    priority: Priority = Priority.FIRST
    entity: Entity

    def __str__(self):
        return f"<RemoveEntityView '{self.entity.name}'>"


class CreateAttribute(Migration):
    priority: Priority = Priority.SECOND
    entity: Entity
    attribute: Attribute

    def __str__(self):
        return f"<CreateAttribute '{self.attribute.name}' entity='{self.entity.name}'>"


class RemoveAttribute(Migration):
    priority: Priority = Priority.SECOND
    entity: Entity
    attribute: Attribute

    def __str__(self):
        return f"<RemoveAttribute '{self.attribute.name}' entity='{self.entity.name}'>"


class CreateGroup(Migration):
    priority: Priority = Priority.SECOND
    entity: Entity
    group: Group

    def __str__(self):
        return f"<CreateGroup '{self.group.name}' entity='{self.entity.name}'>"


class RemoveGroup(Migration):
    priority: Priority = Priority.SECOND
    entity: Entity
    group: Group

    def __str__(self):
        return f"<RemoveGroup '{self.group.name}' entity='{self.entity.name}'>"


class AddGroupAttribute(Migration):
    """Add an Attribute to an existing Group."""

    priority: Priority = Priority.SECOND
    entity: Entity
    group: Group
    attribute: Attribute

    def __str__(self):
        return (
            f"<AddGroupAttribute '{self.group.name}'"
            f" entity='{self.entity.name}'"
            f" attribute='{self.attribute.name}'>"
        )


class RemoveGroupAttribute(Migration):
    """Remove an Attribute from an existing Group."""

    priority: Priority = Priority.SECOND
    entity: Entity
    group: Group
    attribute: Attribute

    def __str__(self):
        return (
            f"<RemoveGroupAttribute '{self.group.name}'"
            f" entity='{self.entity.name}'"
            f" attribute='{self.attribute.name}'>"
        )


class CreateLink(Migration):
    priority: Priority = Priority.SECOND
    link: Link

    def __str__(self):
        return f"<CreateLink '{self.link.name}'>"


class RemoveLink(Migration):
    priority: Priority = Priority.FIRST
    link: Link

    def __str__(self):
        return f"<RemoveLink '{self.link.name}'>"



import importlib

import click

from .core import HnhmError
from .hnhm_registry import HnhmRegistry
from .plan_printer import Color, print_plan


def import_registry(module: str) -> HnhmRegistry:
    if module == ".":
        raise HnhmError(
            "Importing from the current directory '.' is not supported. "
            "Please, use the hnhm from the parent directory."
        )

    module = module.strip("/")
    module = f"{module}.__hnhm__"

    click.secho(
        f"Importing 'registry' object from the module: '{module}'.",
        fg=Color.cyan,
    )
    click.secho()
    imported_module = importlib.import_module(module)

    registry: HnhmRegistry | None = getattr(imported_module, "registry", None)
    if not registry:
        raise HnhmError(
            f"Failed to import registry from module: '{module}'."
            " Please, specify your registry via 'registry' object in your DWH module."
        )
    return registry


@click.group()
def cli():
    pass


@cli.command()
@click.argument("dwh_module")
def plan(dwh_module: str):
    registry = import_registry(dwh_module)

    with registry.hnhm as hnhm:
        plan = hnhm.plan(entities=registry.entities, links=registry.links)
        print_plan(plan)


@cli.command()
@click.argument("dwh_module")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Apply plan without confirmation.",
)
def apply(dwh_module: str, yes: bool):
    registry = import_registry(dwh_module)

    with registry.hnhm as hnhm:
        plan = hnhm.plan(entities=registry.entities, links=registry.links)
        print_plan(plan)

    if plan.is_empty():
        return

    if yes or click.confirm("Apply migrations?"):
        with registry.hnhm as hnhm:
            hnhm.apply(plan)

        click.secho("Applied!", fg="green")



import pydantic

from .entity import Entity
from .layout import Layout


class LinkElement(pydantic.BaseModel):
    """Link element."""

    entity: Entity
    comment: str

    def __str__(self):
        return f"<LinkElement entity='{self.entity.name}'>"

    def __hash__(self):
        return hash(self.entity.name)

    def __eq__(self, other):
        return hash(self) == hash(other)


class Link(pydantic.BaseModel):
    """Link core representation."""

    fqn: str
    name: str
    doc: str
    layout: Layout
    elements: list[LinkElement]
    keys: list[LinkElement]

    def __str__(self):
        return f"<Link '{self.fqn}' layout='{self.layout}'>"



import abc

import pydantic

from .link import Link
from .entity import Entity
from .error import HnhmError


class HnhmStateData(pydantic.BaseModel):
    """State data for hnhm."""

    entities: dict[str, Entity]
    entities_views: set[str]
    links: dict[str, Link]

    def check_entity_exists(self, entity_fqn: str):
        if entity_fqn not in self.entities:
            raise HnhmError(f"Entity '{entity_fqn}' doesn't exist.")

    def check_entity_not_exists(self, entity_fqn: str):
        if entity_fqn in self.entities:
            raise HnhmError(f"Entity '{entity_fqn}' already exists.")

    def check_link_exists(self, link_fqn: str):
        if link_fqn not in self.links:
            raise HnhmError(f"Link '{link_fqn}' doesn't exist.")

    def check_link_not_exists(self, link: str):
        if link in self.links:
            raise HnhmError(f"Link '{link}' already exists.")

    def check_attribute_exists(self, entity_fqn: str, attribute: str):
        self.check_entity_exists(entity_fqn)
        if attribute not in self.entities[entity_fqn].attributes:
            raise HnhmError(f"Attribute '{attribute}' doesn't exist.")

    def check_attribute_not_exists(self, entity_fqn: str, attribute: str):
        self.check_entity_exists(entity_fqn)
        if attribute in self.entities[entity_fqn].attributes:
            raise HnhmError(f"Attribute '{attribute}' already exists.")

    def check_group_exists(self, entity_fqn: str, group: str):
        self.check_entity_exists(entity_fqn)
        if group not in self.entities[entity_fqn].groups:
            raise HnhmError(f"Group '{group}' doesn't exist.")

    def check_group_not_exists(self, entity_fqn: str, group: str):
        self.check_entity_exists(entity_fqn)
        if group in self.entities[entity_fqn].groups:
            raise HnhmError(f"Group '{group}' already exists.")

    def check_group_attribute_exists(self, entity_fqn: str, group: str, attribute: str):
        self.check_group_exists(entity_fqn, group)
        if attribute not in self.entities[entity_fqn].groups[group].attributes:
            raise HnhmError(f"Attribute '{attribute}' doesn't exist.")

    def check_group_attribute_not_exists(
        self, entity_fqn: str, group: str, attribute: str
    ):
        self.check_group_exists(entity_fqn, group)
        if attribute in self.entities[entity_fqn].groups[group].attributes:
            raise HnhmError(f"Attribute '{attribute}' already exists.")


class State(abc.ABC):
    @abc.abstractmethod
    def load(self) -> HnhmStateData:
        raise NotImplementedError

    @abc.abstractmethod
    def save(self, data: HnhmStateData):
        raise NotImplementedError


class InMemState(State):
    def load(self) -> HnhmStateData:
        return HnhmStateData(entities={}, entities_views=set(), links={})

    def save(self, data: HnhmStateData):
        pass



from .hnhm import HnHm
from .core import HnhmError
from .hnhm_link import HnhmLink
from .hnhm_entity import HnhmEntity


class HnhmRegistry:
    def __init__(
        self,
        *,
        hnhm: HnHm,
        entities: list[HnhmEntity] | None = None,
        links: list[HnhmLink] | None = None,
    ):
        entities = entities or []
        links = links or []

        entities_names = set()
        for entity in entities:
            entity_core = entity.to_core()
            full_entity_name = f"{entity_core.layout.type}.{entity_core.name}"
            if full_entity_name in entities_names:
                raise HnhmError(
                    f"Found duplicated entity: '{full_entity_name}'."
                    " Please, use unique name for each entity and LayoutType."
                )
            entities_names.add(full_entity_name)

        links_names = set()
        for link in links:
            link_core = link.to_core()
            if link_core.name in links_names:
                raise HnhmError(
                    f"Found duplicated link: '{link_core.name}'."
                    " Please, use unique name for each link."
                )
            links_names.add(link_core.name)

        self.hnhm = hnhm
        self.entities = entities
        self.links = links

    def __str__(self):
        return "<HnhmRegistry>"



from pathlib import Path

from .core import State, HnhmStateData


class FileState(State):
    def __init__(self, file_name: str):
        self.file_name = file_name

    def load(self) -> HnhmStateData:
        if Path(self.file_name).is_file():
            with open(self.file_name) as file_state:
                return HnhmStateData.model_validate_json(file_state.read())
        else:
            return HnhmStateData(entities={}, entities_views=set(), links={})

    def save(self, data: HnhmStateData):
        with open(self.file_name, "w") as f:
            f.write(data.model_dump_json(indent=2))



from .flow import Flow
from .hnhm import HnHm
from .core import HnhmError
from .file_state import FileState
from .hnhm_registry import HnhmRegistry
from .postgres.sql import PostgresPsycopgSql
from .hnhm_link import HnhmLink, HnhmLinkElement
from .hnhm_entity import Layout, HnhmEntity, LayoutType
from .hnhm_attribute import Float, String, Boolean, Integer, Timestamp, ChangeType

__all__ = [
    "ChangeType",
    "FileState",
    "Float",
    "Flow",
    "HnHm",
    "HnhmEntity",
    "HnhmError",
    "HnhmLink",
    "HnhmLinkElement",
    "HnhmRegistry",
    "Integer",
    "Layout",
    "LayoutType",
    "PostgresPsycopgSql",
    "String",
    "Timestamp",
    "Boolean",
]

__version__ = "0.0.11"



from .sql import Sql
from .group import Group
from .entity import Entity
from .error import HnhmError
from . import task, migration
from .link import Link, LinkElement
from .layout import Layout, LayoutType
from .state import State, HnhmStateData
from .attribute import Attribute, ChangeType

__all__ = [
    "Attribute",
    "ChangeType",
    "Entity",
    "Group",
    "HnhmError",
    "HnhmStateData",
    "Layout",
    "LayoutType",
    "Link",
    "LinkElement",
    "Sql",
    "State",
    "migration",
    "task",
]



from enum import Enum

import click
import pydantic

from .hnhm import Plan, PlanType
from .core import HnhmError, LayoutType, migration


class Color(str, Enum):
    green = "green"
    red = "red"
    yellow = "yellow"
    cyan = "cyan"


class PlanLine(pydantic.BaseModel):
    text: str
    color: str | None = None


def lines_from_plan(plan: Plan) -> list[PlanLine]:
    lines = []
    if plan.is_empty():
        lines.append(PlanLine(text="Your DWH is up to date.", color=Color.green))
        return lines

    entities_migrations = sorted(plan.entities_migrations.items(), key=lambda kv: kv[0])
    links_migrations = sorted(plan.links_migrations.items(), key=lambda kv: kv[0])

    lines.append(PlanLine(text="Plan:"))
    for entity_fqn, plan_collection in entities_migrations:
        if plan_collection.type == PlanType.CREATE:
            symbol, color = "+", Color.green
        elif plan_collection.type == PlanType.REMOVE:
            symbol, color = "-", Color.red
        elif plan_collection.type == PlanType.UPDATE:
            symbol, color = "[u]", Color.yellow
        else:
            raise HnhmError()

        lines.append(PlanLine(text=""))
        lines.append(PlanLine(text=f"{symbol} entity '{entity_fqn}'", color=color))
        for entity_migration in plan_collection.migrations:
            match entity_migration:
                case migration.CreateEntity(entity=entity):
                    if entity.layout.type == LayoutType.HNHM:
                        lines.append(
                            PlanLine(text=f"  + hub '{entity.name}'", color=Color.green)
                        )
                    else:
                        lines.append(
                            PlanLine(text=f"  + stage '{entity.name}'", color=Color.green)
                        )
                        for attribute in entity.attributes.values():
                            lines.append(
                                PlanLine(
                                    text=f"    |attribute '{attribute.name}'",
                                    color=Color.green,
                                )
                            )

                case migration.RecreateEntityView(entity=entity):
                    lines.append(
                        PlanLine(text=f"  {symbol} view '{entity.name}'", color=color)
                    )

                case migration.CreateAttribute(entity=_, attribute=attribute):
                    lines.append(
                        PlanLine(
                            text=f"  + attribute '{attribute.name}'", color=Color.green
                        )
                    )

                case migration.CreateGroup(entity=_, group=group):
                    lines.append(
                        PlanLine(text=f"  + group '{group.name}'", color=Color.green)
                    )
                    for attribute in group.attributes.values():
                        lines.append(
                            PlanLine(
                                text=f"    |attribute '{attribute.name}'",
                                color=Color.green,
                            )
                        )

                case migration.AddGroupAttribute(
                    entity=_, group=group, attribute=attribute
                ):
                    lines.append(
                        PlanLine(text=f"  [u] group '{group.name}'", color=Color.yellow)
                    )
                    lines.append(
                        PlanLine(
                            text=f"    +attribute '{attribute.name}'", color=Color.green
                        )
                    )

                case migration.RemoveEntity(entity=entity):
                    if entity.layout.type == LayoutType.HNHM:
                        lines.append(
                            PlanLine(text=f"  - hub '{entity.name}'", color=Color.red)
                        )
                    else:
                        lines.append(
                            PlanLine(text=f"  - stage '{entity.name}'", color=Color.red)
                        )
                        for attribute in entity.attributes.values():
                            lines.append(
                                PlanLine(
                                    text=f"    |attribute '{attribute.name}'",
                                    color=Color.red,
                                )
                            )

                case migration.RemoveEntityView(entity=entity):
                    if plan_collection.type == PlanType.REMOVE:
                        lines.append(
                            PlanLine(text=f"  - view '{entity.name}'", color=Color.red)
                        )

                case migration.RemoveAttribute(entity=_, attribute=attribute):
                    lines.append(
                        PlanLine(
                            text=f"  - attribute '{attribute.name}'", color=Color.red
                        )
                    )

                case migration.RemoveGroup(entity=_, group=group):
                    lines.append(
                        PlanLine(text=f"  - group '{group.name}'", color=Color.red)
                    )
                    for attribute in group.attributes.values():
                        lines.append(
                            PlanLine(
                                text=f"    | attribute '{attribute.name}'",
                                color=Color.red,
                            )
                        )

                case migration.RemoveGroupAttribute(
                    entity=_, group=group, attribute=attribute
                ):
                    lines.append(
                        PlanLine(text=f"  [u] group '{group.name}'", color=Color.yellow)
                    )
                    lines.append(
                        PlanLine(
                            text=f"    -attribute '{attribute.name}'", color=Color.red
                        )
                    )

    for link_fqn, plan_collection in links_migrations:
        if plan_collection.type == PlanType.CREATE:
            symbol, color = "+", Color.green
        elif plan_collection.type == PlanType.REMOVE:
            symbol, color = "-", Color.red
        else:
            raise HnhmError()

        lines.append(PlanLine(text=""))
        lines.append(PlanLine(text=f"{symbol} link '{link_fqn}'", color=color))
        for link_migration in plan_collection.migrations:
            match link_migration:
                case migration.RemoveLink(link=link):
                    for element in link.elements:
                        lines.append(
                            PlanLine(
                                text=f"  |element '{element.entity.name}'", color=color
                            )
                        )
                case migration.CreateLink(link=link):
                    for element in link.elements:
                        lines.append(
                            PlanLine(
                                text=f"  |element '{element.entity.name}'", color=color
                            )
                        )

    lines.append(PlanLine(text=""))

    return lines


def print_plan(plan: Plan):
    lines = lines_from_plan(plan)
    for line in lines:
        click.secho(line.text, fg=line.color)
